@@ -1,12 +1,11 @@
 use std::os::raw::{c_int, c_void};
-use std::{ptr, slice};
+use std::ptr;
+use std::slice;
 
 use crate::error::{Error, Result};
 use crate::ffi;
 use crate::types::LuaRef;
-use crate::util::{
-    assert_stack, check_stack, error_traceback, pop_error, protect_lua_closure, StackGuard,
-};
+use crate::util::{assert_stack, check_stack, error_traceback, pop_error, protect_lua, StackGuard};
 use crate::value::{FromLuaMulti, MultiValue, ToLuaMulti};
 
 #[cfg(feature = "async")]
@@ -92,7 +91,7 @@ impl<'lua> Function<'lua> {
     /// Returns a Feature that, when polled, calls `self`, passing `args` as function arguments,
     /// and drives the execution.
     ///
-    /// Internaly it wraps the function to an [`AsyncThread`].
+    /// Internally it wraps the function to an [`AsyncThread`].
     ///
     /// Requires `feature = "async"`
     ///
@@ -193,13 +192,13 @@ impl<'lua> Function<'lua> {
         unsafe {
             let _sg = StackGuard::new(lua.state);
             check_stack(lua.state, nargs + 5)?;
+
             lua.push_ref(&self.0);
             ffi::lua_pushinteger(lua.state, nargs as ffi::lua_Integer);
             for arg in args {
                 lua.push_value(arg)?;
             }
-
-            protect_lua_closure(lua.state, nargs + 2, 1, |state| {
+            protect_lua(lua.state, nargs + 2, 1, |state| {
                 ffi::lua_pushcclosure(state, bind_call_impl, nargs + 2);
             })?;
 
@@ -211,7 +210,7 @@ impl<'lua> Function<'lua> {
     ///
     /// If `strip` is true, the binary representation may not include all debug information
     /// about the function, to save space.
-    pub fn dump(&self, strip: bool) -> Result<Vec<u8>> {
+    pub fn dump(&self, strip: bool) -> Vec<u8> {
         unsafe extern "C" fn writer(
             _state: *mut ffi::lua_State,
             buf: *const c_void,
@@ -229,18 +228,15 @@ impl<'lua> Function<'lua> {
         unsafe {
             let _sg = StackGuard::new(lua.state);
             assert_stack(lua.state, 1);
+
             lua.push_ref(&self.0);
+            let data_ptr = &mut data as *mut Vec<u8> as *mut c_void;
             let strip = if strip { 1 } else { 0 };
-            ffi::lua_dump(
-                lua.state,
-                writer,
-                &mut data as *mut Vec<u8> as *mut c_void,
-                strip,
-            );
+            ffi::lua_dump(lua.state, writer, data_ptr, strip);
             ffi::lua_pop(lua.state, 1);
         }
 
-        Ok(data)
+        data
     }
 }
 
