@@ -9,9 +9,9 @@ use crate::util::callback_error;
 /// Contains information about currently executing Lua code.
 ///
 /// The `Debug` structure is provided as a parameter to the hook function set with
-/// [`Lua::set_hook`].  You may call the methods on this structure to retrieve information about the
-/// Lua code executing at the time that the hook function was called.  Further information can be
-/// found in the [Lua 5.3 documentaton][lua_doc].
+/// [`Lua::set_hook`]. You may call the methods on this structure to retrieve information about the
+/// Lua code executing at the time that the hook function was called. Further information can be
+/// found in the [Lua 5.3 documentation][lua_doc].
 ///
 /// [lua_doc]: https://www.lua.org/manual/5.3/manual.html#lua_Debug
 /// [`Lua::set_hook`]: struct.Lua.html#method.set_hook
@@ -130,7 +130,7 @@ pub struct HookTriggers {
     pub on_returns: bool,
     /// Before executing a new line, or returning from a function call.
     pub every_line: bool,
-    /// After a certain number of VM instructions have been executed.  When set to `Some(count)`,
+    /// After a certain number of VM instructions have been executed. When set to `Some(count)`,
     /// `count` is the number of VM instructions to execute before calling the hook.
     ///
     /// # Performance
@@ -173,7 +173,7 @@ pub(crate) unsafe extern "C" fn hook_proc(state: *mut lua_State, ar: *mut lua_De
             _phantom: PhantomData,
         };
 
-        let lua = Lua::make_from_ptr(state);
+        let lua = mlua_expect!(Lua::make_from_ptr(state), "cannot make Lua instance");
         let hook_cb = mlua_expect!(lua.hook_callback(), "no hook callback set in hook_proc");
 
         #[allow(clippy::match_wild_err_arm)]
@@ -183,7 +183,7 @@ pub(crate) unsafe extern "C" fn hook_proc(state: *mut lua_State, ar: *mut lua_De
         }?;
 
         Ok(())
-    });
+    })
 }
 
 unsafe fn ptr_to_str<'a>(input: *const c_char) -> Option<&'a [u8]> {
