@@ -1,6 +1,6 @@
 //! (De)Serialization support using serde.
 
-use std::os::raw::{c_int, c_void};
+use std::os::raw::c_void;
 use std::ptr;
 
 use serde::{Deserialize, Serialize};
@@ -9,9 +9,11 @@ use crate::error::Result;
 use crate::ffi;
 use crate::lua::Lua;
 use crate::table::Table;
-use crate::util::{assert_stack, protect_lua, StackGuard};
+use crate::types::LightUserData;
+use crate::util::{assert_stack, check_stack, protect_lua, StackGuard};
 use crate::value::Value;
 
+/// Trait for serializing/deserializing Lua values using Serde.
 pub trait LuaSerdeExt<'lua> {
     /// A special value (lightuserdata) to encode/decode optional (none) values.
     ///
@@ -25,7 +27,7 @@ pub trait LuaSerdeExt<'lua> {
     ///
     /// fn main() -> Result<()> {
     ///     let lua = Lua::new();
-    ///     lua.globals().set("null", lua.null()?)?;
+    ///     lua.globals().set("null", lua.null())?;
     ///
     ///     let val = lua.load(r#"{a = null}"#).eval()?;
     ///     let map: HashMap<String, Option<String>> = lua.from_value(val)?;
@@ -34,7 +36,7 @@ pub trait LuaSerdeExt<'lua> {
     ///     Ok(())
     /// }
     /// ```
-    fn null(&'lua self) -> Result<Value<'lua>>;
+    fn null(&'lua self) -> Value<'lua>;
 
     /// A metatable attachable to a Lua table to systematically encode it as Array (instead of Map).
     /// As result, encoded Array will contain only sequence part of the table, with the same length
@@ -50,7 +52,7 @@ pub trait LuaSerdeExt<'lua> {
     ///
     /// fn main() -> Result<()> {
     ///     let lua = Lua::new();
-    ///     lua.globals().set("array_mt", lua.array_metatable()?)?;
+    ///     lua.globals().set("array_mt", lua.array_metatable())?;
     ///
     ///     // Encode as an empty array (no sequence part in the lua table)
     ///     let val = lua.load("setmetatable({a = 5}, array_mt)").eval()?;
@@ -65,7 +67,7 @@ pub trait LuaSerdeExt<'lua> {
     ///     Ok(())
     /// }
     /// ```
-    fn array_metatable(&'lua self) -> Result<Table<'lua>>;
+    fn array_metatable(&'lua self) -> Table<'lua>;
 
     /// Converts `T` into a `Value` instance.
     ///
@@ -100,6 +102,33 @@ pub trait LuaSerdeExt<'lua> {
     /// ```
     fn to_value<T: Serialize + ?Sized>(&'lua self, t: &T) -> Result<Value<'lua>>;
 
+    /// Converts `T` into a `Value` instance with options.
+    ///
+    /// Requires `feature = "serialize"`
+    ///
+    /// [`Value`]: enum.Value.html
+    ///
+    /// # Example
+    ///
+    /// ```
+    /// use mlua::{Lua, Result, LuaSerdeExt, SerializeOptions};
+    ///
+    /// fn main() -> Result<()> {
+    ///     let lua = Lua::new();
+    ///     let v = vec![1, 2, 3];
+    ///     let options = SerializeOptions::new().set_array_metatable(false);
+    ///     lua.globals().set("v", lua.to_value_with(&v, options)?)?;
+    ///
+    ///     lua.load(r#"
+    ///         assert(#v == 3 and v[1] == 1 and v[2] == 2 and v[3] == 3)
+    ///         assert(getmetatable(v) == nil)
+    ///     "#).exec()
+    /// }
+    /// ```
+    fn to_value_with<T>(&'lua self, t: &T, options: ser::Options) -> Result<Value<'lua>>
+    where
+        T: Serialize + ?Sized;
+
     /// Deserializes a `Value` into any serde deserializable object.
     ///
     /// Requires `feature = "serialize"`
@@ -132,31 +161,18 @@ pub trait LuaSerdeExt<'lua> {
 }
 
 impl<'lua> LuaSerdeExt<'lua> for Lua {
-    fn null(&'lua self) -> Result<Value<'lua>> {
-        unsafe {
-            let _sg = StackGuard::new(self.state);
-            assert_stack(self.state, 3);
-
-            unsafe extern "C" fn push_null(state: *mut ffi::lua_State) -> c_int {
-                ffi::lua_pushlightuserdata(state, ptr::null_mut());
-                1
-            }
-            protect_lua(self.state, 0, push_null)?;
-            Ok(self.pop_value())
-        }
+    fn null(&'lua self) -> Value<'lua> {
+        Value::LightUserData(LightUserData(ptr::null_mut()))
     }
 
-    fn array_metatable(&'lua self) -> Result<Table<'lua>> {
+    fn array_metatable(&'lua self) -> Table<'lua> {
         unsafe {
             let _sg = StackGuard::new(self.state);
-            assert_stack(self.state, 3);
+            assert_stack(self.state, 1);
 
-            unsafe extern "C" fn get_array_mt(state: *mut ffi::lua_State) -> c_int {
-                push_array_metatable(state);
-                1
-            }
-            protect_lua(self.state, 0, get_array_mt)?;
-            Ok(Table(self.pop_ref()))
+            push_array_metatable(self.state);
+
+            Table(self.pop_ref())
         }
     }
 
@@ -164,37 +180,50 @@ impl<'lua> LuaSerdeExt<'lua> for Lua {
     where
         T: Serialize + ?Sized,
     {
-        t.serialize(ser::Serializer(self))
+        t.serialize(ser::Serializer::new(self))
+    }
+
+    fn to_value_with<T>(&'lua self, t: &T, options: ser::Options) -> Result<Value<'lua>>
+    where
+        T: Serialize + ?Sized,
+    {
+        t.serialize(ser::Serializer::new_with_options(self, options))
     }
 
     fn from_value<T>(&'lua self, value: Value<'lua>) -> Result<T>
     where
         T: Deserialize<'lua>,
     {
-        T::deserialize(de::Deserializer(value))
+        T::deserialize(de::Deserializer::new(value))
     }
 }
 
-pub(crate) unsafe fn init_metatables(state: *mut ffi::lua_State) {
-    ffi::lua_pushlightuserdata(
-        state,
-        &ARRAY_METATABLE_REGISTRY_KEY as *const u8 as *mut c_void,
-    );
-    ffi::lua_newtable(state);
+// Uses 6 stack spaces and calls checkstack.
+pub(crate) unsafe fn init_metatables(state: *mut ffi::lua_State) -> Result<()> {
+    check_stack(state, 3)?;
+    protect_lua(state, 0, 0, |state| {
+        ffi::lua_createtable(state, 0, 1);
 
-    ffi::lua_pushstring(state, cstr!("__metatable"));
-    ffi::lua_pushboolean(state, 0);
-    ffi::lua_rawset(state, -3);
+        ffi::lua_pushstring(state, cstr!("__metatable"));
+        ffi::lua_pushboolean(state, 0);
+        ffi::lua_rawset(state, -3);
 
-    ffi::lua_rawset(state, ffi::LUA_REGISTRYINDEX);
+        let array_metatable_key = &ARRAY_METATABLE_REGISTRY_KEY as *const u8 as *const c_void;
+        ffi::lua_rawsetp(state, ffi::LUA_REGISTRYINDEX, array_metatable_key);
+    })
 }
 
 pub(crate) unsafe fn push_array_metatable(state: *mut ffi::lua_State) {
-    let key = &ARRAY_METATABLE_REGISTRY_KEY as *const u8 as *mut c_void;
-    ffi::lua_rawgetp(state, ffi::LUA_REGISTRYINDEX, key);
+    let array_metatable_key = &ARRAY_METATABLE_REGISTRY_KEY as *const u8 as *mut c_void;
+    ffi::lua_rawgetp(state, ffi::LUA_REGISTRYINDEX, array_metatable_key);
 }
 
 static ARRAY_METATABLE_REGISTRY_KEY: u8 = 0;
 
 pub mod de;
 pub mod ser;
+
+#[doc(inline)]
+pub use de::Deserializer;
+#[doc(inline)]
+pub use ser::Serializer;
