@@ -1,25 +1,27 @@
+use std::cmp;
 use std::os::raw::c_int;
 
 use crate::error::{Error, Result};
 use crate::ffi;
 use crate::types::LuaRef;
-use crate::util::{
-    assert_stack, check_stack, error_traceback, pop_error, protect_lua_closure, StackGuard,
-};
+use crate::util::{assert_stack, check_stack, error_traceback, pop_error, protect_lua, StackGuard};
 use crate::value::{FromLuaMulti, MultiValue, ToLuaMulti};
 
+#[cfg(any(feature = "lua54", all(feature = "luajit", feature = "vendored"), doc))]
+use crate::function::Function;
+
 #[cfg(feature = "async")]
 use {
     crate::{
-        error::ExternalError,
-        lua::{AsyncPollPending, Lua, WAKER_REGISTRY_KEY},
-        util::{get_gc_userdata, push_gc_userdata},
+        lua::{ASYNC_POLL_PENDING, WAKER_REGISTRY_KEY},
+        util::get_gc_userdata,
         value::Value,
     },
     futures_core::{future::Future, stream::Stream},
     std::{
         cell::RefCell,
         marker::PhantomData,
+        mem,
         os::raw::c_void,
         pin::Pin,
         task::{Context, Poll, Waker},
@@ -109,24 +111,21 @@ impl<'lua> Thread<'lua> {
     {
         let lua = self.0.lua;
         let args = args.to_lua_multi(lua)?;
+        let nargs = args.len() as c_int;
         let results = unsafe {
             let _sg = StackGuard::new(lua.state);
-            assert_stack(lua.state, 3);
+            check_stack(lua.state, cmp::min(nargs + 1, 3))?;
 
             lua.push_ref(&self.0);
             let thread_state = ffi::lua_tothread(lua.state, -1);
+            ffi::lua_pop(lua.state, 1);
 
             let status = ffi::lua_status(thread_state);
             if status != ffi::LUA_YIELD && ffi::lua_gettop(thread_state) == 0 {
                 return Err(Error::CoroutineInactive);
             }
 
-            ffi::lua_pop(lua.state, 1);
-
-            let nargs = args.len() as c_int;
-            check_stack(lua.state, nargs)?;
-            check_stack(thread_state, nargs + 1)?;
-
+            check_stack(thread_state, nargs)?;
             for arg in args {
                 lua.push_value(arg)?;
             }
@@ -136,17 +135,14 @@ impl<'lua> Thread<'lua> {
 
             let ret = ffi::lua_resume(thread_state, lua.state, nargs, &mut nresults as *mut c_int);
             if ret != ffi::LUA_OK && ret != ffi::LUA_YIELD {
-                protect_lua_closure(lua.state, 0, 0, |_| {
-                    error_traceback(thread_state);
-                    0
-                })?;
+                protect_lua(lua.state, 0, 0, |_| error_traceback(thread_state))?;
                 return Err(pop_error(thread_state, ret));
             }
 
             let mut results = MultiValue::new();
+            check_stack(lua.state, nresults + 2)?; // 2 is extra for `lua.pop_value()` below
             ffi::lua_xmove(thread_state, lua.state, nresults);
 
-            assert_stack(lua.state, 2);
             for _ in 0..nresults {
                 results.push_front(lua.pop_value());
             }
@@ -177,6 +173,43 @@ impl<'lua> Thread<'lua> {
         }
     }
 
+    /// Resets a thread
+    ///
+    /// In [Lua 5.4]: cleans its call stack and closes all pending to-be-closed variables.
+    /// Returns a error in case of either the original error that stopped the thread or errors
+    /// in closing methods.
+    ///
+    /// In [LuaJIT]: resets to the initial state of a newly created Lua thread.
+    /// Lua threads in arbitrary states (like yielded or errored) can be reset properly.
+    ///
+    /// Sets a Lua function for the thread afterwards.
+    ///
+    /// Requires `feature = "lua54"` OR `feature = "luajit,vendored"`
+    ///
+    /// [Lua 5.4]: https://www.lua.org/manual/5.4/manual.html#lua_resetthread
+    /// [LuaJIT]: https://github.com/openresty/luajit2#lua_resetthread
+    #[cfg(any(feature = "lua54", all(feature = "luajit", feature = "vendored"), doc))]
+    pub fn reset(&self, func: Function<'lua>) -> Result<()> {
+        let lua = self.0.lua;
+        unsafe {
+            let _sg = StackGuard::new(lua.state);
+            check_stack(lua.state, 2)?;
+
+            lua.push_ref(&self.0);
+            let thread_state = ffi::lua_tothread(lua.state, -1);
+
+            let ret = ffi::lua_resetthread(lua.state, thread_state);
+            if ret != ffi::LUA_OK {
+                return Err(pop_error(thread_state, ret));
+            }
+
+            lua.push_ref(&func.0);
+            ffi::lua_xmove(lua.state, thread_state, 1);
+
+            Ok(())
+        }
+    }
+
     /// Converts Thread to an AsyncThread which implements Future and Stream traits.
     ///
     /// `args` are passed as arguments to the thread function for first call.
@@ -225,7 +258,7 @@ impl<'lua> Thread<'lua> {
         A: ToLuaMulti<'lua>,
         R: FromLuaMulti<'lua>,
     {
-        let args = args.to_lua_multi(&self.0.lua);
+        let args = args.to_lua_multi(self.0.lua);
         AsyncThread {
             thread: self,
             args0: RefCell::new(Some(args)),
@@ -262,7 +295,7 @@ where
             self.thread.resume(())?
         };
 
-        if is_poll_pending(lua, &ret) {
+        if is_poll_pending(&ret) {
             return Poll::Pending;
         }
 
@@ -283,7 +316,7 @@ where
 
         match self.thread.status() {
             ThreadStatus::Resumable => {}
-            _ => return Poll::Ready(Err("Thread already finished".to_lua_err())),
+            _ => return Poll::Ready(Err(Error::CoroutineInactive)),
         };
 
         let _wg = WakerGuard::new(lua.state, cx.waker().clone());
@@ -293,7 +326,7 @@ where
             self.thread.resume(())?
         };
 
-        if is_poll_pending(lua, &ret) {
+        if is_poll_pending(&ret) {
             return Poll::Pending;
         }
 
@@ -308,44 +341,31 @@ where
 }
 
 #[cfg(feature = "async")]
-fn is_poll_pending(lua: &Lua, val: &MultiValue) -> bool {
-    if val.len() != 1 {
-        return false;
-    }
-
-    if let Some(Value::UserData(ud)) = val.iter().next() {
-        unsafe {
-            let _sg = StackGuard::new(lua.state);
-            assert_stack(lua.state, 3);
-
-            lua.push_ref(&ud.0);
-            let is_pending = get_gc_userdata::<AsyncPollPending>(lua.state, -1)
-                .as_ref()
-                .is_some();
-            ffi::lua_pop(lua.state, 1);
-
-            return is_pending;
+fn is_poll_pending(val: &MultiValue) -> bool {
+    match val.iter().enumerate().last() {
+        Some((1, Value::LightUserData(ud))) => {
+            ud.0 == &ASYNC_POLL_PENDING as *const u8 as *mut c_void
         }
+        _ => false,
     }
-
-    false
 }
 
 #[cfg(feature = "async")]
-struct WakerGuard(*mut ffi::lua_State);
+struct WakerGuard(*mut ffi::lua_State, Option<Waker>);
 
 #[cfg(feature = "async")]
 impl WakerGuard {
     pub fn new(state: *mut ffi::lua_State, waker: Waker) -> Result<WakerGuard> {
         unsafe {
             let _sg = StackGuard::new(state);
-            assert_stack(state, 6);
+            check_stack(state, 3)?;
 
-            ffi::lua_pushlightuserdata(state, &WAKER_REGISTRY_KEY as *const u8 as *mut c_void);
-            push_gc_userdata(state, waker)?;
-            ffi::lua_rawset(state, ffi::LUA_REGISTRYINDEX);
+            let waker_key = &WAKER_REGISTRY_KEY as *const u8 as *const c_void;
+            ffi::lua_rawgetp(state, ffi::LUA_REGISTRYINDEX, waker_key);
+            let waker_slot = get_gc_userdata::<Option<Waker>>(state, -1).as_mut();
+            let old = mlua_expect!(waker_slot, "Waker is destroyed").replace(waker);
 
-            Ok(WakerGuard(state))
+            Ok(WakerGuard(state, old))
         }
     }
 }
@@ -353,14 +373,15 @@ impl WakerGuard {
 #[cfg(feature = "async")]
 impl Drop for WakerGuard {
     fn drop(&mut self) {
+        let state = self.0;
         unsafe {
-            let state = self.0;
             let _sg = StackGuard::new(state);
-            assert_stack(state, 2);
+            assert_stack(state, 3);
 
-            ffi::lua_pushlightuserdata(state, &WAKER_REGISTRY_KEY as *const u8 as *mut c_void);
-            ffi::lua_pushnil(state);
-            ffi::lua_rawset(state, ffi::LUA_REGISTRYINDEX);
+            let waker_key = &WAKER_REGISTRY_KEY as *const u8 as *const c_void;
+            ffi::lua_rawgetp(state, ffi::LUA_REGISTRYINDEX, waker_key);
+            let waker_slot = get_gc_userdata::<Option<Waker>>(state, -1).as_mut();
+            mem::swap(mlua_expect!(waker_slot, "Waker is destroyed"), &mut self.1);
         }
     }
 }
