@@ -1,4 +1,9 @@
-use std::cell::{Ref, RefMut};
+use std::any::TypeId;
+use std::cell::{Ref, RefCell, RefMut};
+use std::fmt;
+use std::hash::{Hash, Hasher};
+use std::ops::{Deref, DerefMut};
+use std::string::String as StdString;
 
 #[cfg(feature = "async")]
 use std::future::Future;
@@ -13,10 +18,18 @@ use crate::error::{Error, Result};
 use crate::ffi;
 use crate::function::Function;
 use crate::lua::Lua;
-use crate::table::Table;
-use crate::types::{LuaRef, MaybeSend, UserDataCell};
-use crate::util::{assert_stack, get_destructed_userdata_metatable, get_userdata, StackGuard};
-use crate::value::{FromLua, FromLuaMulti, ToLua, ToLuaMulti, Value};
+use crate::table::{Table, TablePairs};
+use crate::types::{Callback, LuaRef, MaybeSend};
+use crate::util::{
+    check_stack, get_destructed_userdata_metatable, get_userdata, push_string, StackGuard,
+};
+use crate::value::{FromLua, FromLuaMulti, ToLua, ToLuaMulti};
+
+#[cfg(any(feature = "lua52", feature = "lua51", feature = "luajit"))]
+use crate::value::Value;
+
+#[cfg(feature = "async")]
+use crate::types::AsyncCallback;
 
 /// Kinds of metamethods that can be overridden.
 ///
@@ -24,7 +37,7 @@ use crate::value::{FromLua, FromLuaMulti, ToLua, ToLuaMulti, Value};
 /// generally no need to do so: [`UserData`] implementors can instead just implement `Drop`.
 ///
 /// [`UserData`]: trait.UserData.html
-#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
+#[derive(Debug, Clone)]
 pub enum MetaMethod {
     /// The `+` operator.
     Add,
@@ -93,6 +106,15 @@ pub enum MetaMethod {
     /// Requires `feature = "lua54/lua53/lua52"`
     #[cfg(any(feature = "lua54", feature = "lua53", feature = "lua52", doc))]
     Pairs,
+    /// The `__ipairs` metamethod.
+    ///
+    /// This is not an operator, but it will be called by the built-in [`ipairs`] function.
+    ///
+    /// Requires `feature = "lua52"`
+    ///
+    /// [`ipairs`]: https://www.lua.org/manual/5.2/manual.html#pdf-ipairs
+    #[cfg(any(feature = "lua52", doc))]
+    IPairs,
     /// The `__close` metamethod.
     ///
     /// Executed when a variable, that marked as to-be-closed, goes out of scope.
@@ -105,58 +127,155 @@ pub enum MetaMethod {
     /// [lua_doc]: https://www.lua.org/manual/5.4/manual.html#3.3.8
     #[cfg(any(feature = "lua54", doc))]
     Close,
+    /// A custom metamethod.
+    ///
+    /// Must not be in the protected list: `__gc`, `__metatable`, `__mlua*`.
+    Custom(StdString),
+}
+
+impl PartialEq for MetaMethod {
+    fn eq(&self, other: &Self) -> bool {
+        self.name() == other.name()
+    }
+}
+
+impl Eq for MetaMethod {}
+
+impl Hash for MetaMethod {
+    fn hash<H: Hasher>(&self, state: &mut H) {
+        self.name().hash(state);
+    }
+}
+
+impl fmt::Display for MetaMethod {
+    fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(fmt, "{}", self.name())
+    }
 }
 
 impl MetaMethod {
-    pub(crate) fn name(self) -> &'static [u8] {
+    /// Returns Lua metamethod name, usually prefixed by two underscores.
+    pub fn name(&self) -> &str {
         match self {
-            MetaMethod::Add => b"__add",
-            MetaMethod::Sub => b"__sub",
-            MetaMethod::Mul => b"__mul",
-            MetaMethod::Div => b"__div",
-            MetaMethod::Mod => b"__mod",
-            MetaMethod::Pow => b"__pow",
-            MetaMethod::Unm => b"__unm",
+            MetaMethod::Add => "__add",
+            MetaMethod::Sub => "__sub",
+            MetaMethod::Mul => "__mul",
+            MetaMethod::Div => "__div",
+            MetaMethod::Mod => "__mod",
+            MetaMethod::Pow => "__pow",
+            MetaMethod::Unm => "__unm",
 
             #[cfg(any(feature = "lua54", feature = "lua53"))]
-            MetaMethod::IDiv => b"__idiv",
+            MetaMethod::IDiv => "__idiv",
             #[cfg(any(feature = "lua54", feature = "lua53"))]
-            MetaMethod::BAnd => b"__band",
+            MetaMethod::BAnd => "__band",
             #[cfg(any(feature = "lua54", feature = "lua53"))]
-            MetaMethod::BOr => b"__bor",
+            MetaMethod::BOr => "__bor",
             #[cfg(any(feature = "lua54", feature = "lua53"))]
-            MetaMethod::BXor => b"__bxor",
+            MetaMethod::BXor => "__bxor",
             #[cfg(any(feature = "lua54", feature = "lua53"))]
-            MetaMethod::BNot => b"__bnot",
+            MetaMethod::BNot => "__bnot",
             #[cfg(any(feature = "lua54", feature = "lua53"))]
-            MetaMethod::Shl => b"__shl",
+            MetaMethod::Shl => "__shl",
             #[cfg(any(feature = "lua54", feature = "lua53"))]
-            MetaMethod::Shr => b"__shr",
+            MetaMethod::Shr => "__shr",
 
-            MetaMethod::Concat => b"__concat",
-            MetaMethod::Len => b"__len",
-            MetaMethod::Eq => b"__eq",
-            MetaMethod::Lt => b"__lt",
-            MetaMethod::Le => b"__le",
-            MetaMethod::Index => b"__index",
-            MetaMethod::NewIndex => b"__newindex",
-            MetaMethod::Call => b"__call",
-            MetaMethod::ToString => b"__tostring",
+            MetaMethod::Concat => "__concat",
+            MetaMethod::Len => "__len",
+            MetaMethod::Eq => "__eq",
+            MetaMethod::Lt => "__lt",
+            MetaMethod::Le => "__le",
+            MetaMethod::Index => "__index",
+            MetaMethod::NewIndex => "__newindex",
+            MetaMethod::Call => "__call",
+            MetaMethod::ToString => "__tostring",
 
             #[cfg(any(feature = "lua54", feature = "lua53", feature = "lua52"))]
-            MetaMethod::Pairs => b"__pairs",
+            MetaMethod::Pairs => "__pairs",
+            #[cfg(feature = "lua52")]
+            MetaMethod::IPairs => "__ipairs",
 
             #[cfg(feature = "lua54")]
-            MetaMethod::Close => b"__close",
+            MetaMethod::Close => "__close",
+
+            MetaMethod::Custom(ref name) => name,
         }
     }
+
+    pub(crate) fn validate(self) -> Result<Self> {
+        match self {
+            MetaMethod::Custom(name) if name == "__gc" => Err(Error::MetaMethodRestricted(name)),
+            MetaMethod::Custom(name) if name == "__metatable" => {
+                Err(Error::MetaMethodRestricted(name))
+            }
+            MetaMethod::Custom(name) if name.starts_with("__mlua") => {
+                Err(Error::MetaMethodRestricted(name))
+            }
+            _ => Ok(self),
+        }
+    }
+}
+
+impl From<StdString> for MetaMethod {
+    fn from(name: StdString) -> Self {
+        match name.as_str() {
+            "__add" => MetaMethod::Add,
+            "__sub" => MetaMethod::Sub,
+            "__mul" => MetaMethod::Mul,
+            "__div" => MetaMethod::Div,
+            "__mod" => MetaMethod::Mod,
+            "__pow" => MetaMethod::Pow,
+            "__unm" => MetaMethod::Unm,
+
+            #[cfg(any(feature = "lua54", feature = "lua53"))]
+            "__idiv" => MetaMethod::IDiv,
+            #[cfg(any(feature = "lua54", feature = "lua53"))]
+            "__band" => MetaMethod::BAnd,
+            #[cfg(any(feature = "lua54", feature = "lua53"))]
+            "__bor" => MetaMethod::BOr,
+            #[cfg(any(feature = "lua54", feature = "lua53"))]
+            "__bxor" => MetaMethod::BXor,
+            #[cfg(any(feature = "lua54", feature = "lua53"))]
+            "__bnot" => MetaMethod::BNot,
+            #[cfg(any(feature = "lua54", feature = "lua53"))]
+            "__shl" => MetaMethod::Shl,
+            #[cfg(any(feature = "lua54", feature = "lua53"))]
+            "__shr" => MetaMethod::Shr,
+
+            "__concat" => MetaMethod::Concat,
+            "__len" => MetaMethod::Len,
+            "__eq" => MetaMethod::Eq,
+            "__lt" => MetaMethod::Lt,
+            "__le" => MetaMethod::Le,
+            "__index" => MetaMethod::Index,
+            "__newindex" => MetaMethod::NewIndex,
+            "__call" => MetaMethod::Call,
+            "__tostring" => MetaMethod::ToString,
+
+            #[cfg(any(feature = "lua54", feature = "lua53", feature = "lua52"))]
+            "__pairs" => MetaMethod::Pairs,
+            #[cfg(feature = "lua52")]
+            "__ipairs" => MetaMethod::IPairs,
+
+            #[cfg(feature = "lua54")]
+            "__close" => MetaMethod::Close,
+
+            _ => MetaMethod::Custom(name),
+        }
+    }
+}
+
+impl From<&str> for MetaMethod {
+    fn from(name: &str) -> Self {
+        MetaMethod::from(name.to_owned())
+    }
 }
 
 /// Method registry for [`UserData`] implementors.
 ///
 /// [`UserData`]: trait.UserData.html
 pub trait UserDataMethods<'lua, T: UserData> {
-    /// Add a method which accepts a `&T` as the first parameter.
+    /// Add a regular method which accepts a `&T` as the first parameter.
     ///
     /// Regular methods are implemented by overriding the `__index` metamethod and returning the
     /// accessed method. This allows them to be used with the expected `userdata:method()` syntax.
@@ -165,7 +284,7 @@ pub trait UserDataMethods<'lua, T: UserData> {
     /// be used as a fall-back if no regular method is found.
     fn add_method<S, A, R, M>(&mut self, name: &S, method: M)
     where
-        S: ?Sized + AsRef<[u8]>,
+        S: AsRef<[u8]> + ?Sized,
         A: FromLuaMulti<'lua>,
         R: ToLuaMulti<'lua>,
         M: 'static + MaybeSend + Fn(&'lua Lua, &T, A) -> Result<R>;
@@ -177,7 +296,7 @@ pub trait UserDataMethods<'lua, T: UserData> {
     /// [`add_method`]: #method.add_method
     fn add_method_mut<S, A, R, M>(&mut self, name: &S, method: M)
     where
-        S: ?Sized + AsRef<[u8]>,
+        S: AsRef<[u8]> + ?Sized,
         A: FromLuaMulti<'lua>,
         R: ToLuaMulti<'lua>,
         M: 'static + MaybeSend + FnMut(&'lua Lua, &mut T, A) -> Result<R>;
@@ -195,24 +314,25 @@ pub trait UserDataMethods<'lua, T: UserData> {
     fn add_async_method<S, A, R, M, MR>(&mut self, name: &S, method: M)
     where
         T: Clone,
-        S: ?Sized + AsRef<[u8]>,
+        S: AsRef<[u8]> + ?Sized,
         A: FromLuaMulti<'lua>,
         R: ToLuaMulti<'lua>,
         M: 'static + MaybeSend + Fn(&'lua Lua, T, A) -> MR,
         MR: 'lua + Future<Output = Result<R>>;
 
     /// Add a regular method as a function which accepts generic arguments, the first argument will
-    /// be a `UserData` of type T if the method is called with Lua method syntax:
+    /// be a [`AnyUserData`] of type `T` if the method is called with Lua method syntax:
     /// `my_userdata:my_method(arg1, arg2)`, or it is passed in as the first argument:
     /// `my_userdata.my_method(my_userdata, arg1, arg2)`.
     ///
     /// Prefer to use [`add_method`] or [`add_method_mut`] as they are easier to use.
     ///
+    /// [`AnyUserData`]: struct.AnyUserData.html
     /// [`add_method`]: #method.add_method
     /// [`add_method_mut`]: #method.add_method_mut
     fn add_function<S, A, R, F>(&mut self, name: &S, function: F)
     where
-        S: ?Sized + AsRef<[u8]>,
+        S: AsRef<[u8]> + ?Sized,
         A: FromLuaMulti<'lua>,
         R: ToLuaMulti<'lua>,
         F: 'static + MaybeSend + Fn(&'lua Lua, A) -> Result<R>;
@@ -224,7 +344,7 @@ pub trait UserDataMethods<'lua, T: UserData> {
     /// [`add_function`]: #method.add_function
     fn add_function_mut<S, A, R, F>(&mut self, name: &S, function: F)
     where
-        S: ?Sized + AsRef<[u8]>,
+        S: AsRef<[u8]> + ?Sized,
         A: FromLuaMulti<'lua>,
         R: ToLuaMulti<'lua>,
         F: 'static + MaybeSend + FnMut(&'lua Lua, A) -> Result<R>;
@@ -241,8 +361,7 @@ pub trait UserDataMethods<'lua, T: UserData> {
     #[cfg_attr(docsrs, doc(cfg(feature = "async")))]
     fn add_async_function<S, A, R, F, FR>(&mut self, name: &S, function: F)
     where
-        T: Clone,
-        S: ?Sized + AsRef<[u8]>,
+        S: AsRef<[u8]> + ?Sized,
         A: FromLuaMulti<'lua>,
         R: ToLuaMulti<'lua>,
         F: 'static + MaybeSend + Fn(&'lua Lua, A) -> FR,
@@ -256,8 +375,9 @@ pub trait UserDataMethods<'lua, T: UserData> {
     /// side has a metatable. To prevent this, use [`add_meta_function`].
     ///
     /// [`add_meta_function`]: #method.add_meta_function
-    fn add_meta_method<A, R, M>(&mut self, meta: MetaMethod, method: M)
+    fn add_meta_method<S, A, R, M>(&mut self, meta: S, method: M)
     where
+        S: Into<MetaMethod>,
         A: FromLuaMulti<'lua>,
         R: ToLuaMulti<'lua>,
         M: 'static + MaybeSend + Fn(&'lua Lua, &T, A) -> Result<R>;
@@ -270,8 +390,9 @@ pub trait UserDataMethods<'lua, T: UserData> {
     /// side has a metatable. To prevent this, use [`add_meta_function`].
     ///
     /// [`add_meta_function`]: #method.add_meta_function
-    fn add_meta_method_mut<A, R, M>(&mut self, meta: MetaMethod, method: M)
+    fn add_meta_method_mut<S, A, R, M>(&mut self, meta: S, method: M)
     where
+        S: Into<MetaMethod>,
         A: FromLuaMulti<'lua>,
         R: ToLuaMulti<'lua>,
         M: 'static + MaybeSend + FnMut(&'lua Lua, &mut T, A) -> Result<R>;
@@ -281,8 +402,9 @@ pub trait UserDataMethods<'lua, T: UserData> {
     /// Metamethods for binary operators can be triggered if either the left or right argument to
     /// the binary operator has a metatable, so the first argument here is not necessarily a
     /// userdata of type `T`.
-    fn add_meta_function<A, R, F>(&mut self, meta: MetaMethod, function: F)
+    fn add_meta_function<S, A, R, F>(&mut self, meta: S, function: F)
     where
+        S: Into<MetaMethod>,
         A: FromLuaMulti<'lua>,
         R: ToLuaMulti<'lua>,
         F: 'static + MaybeSend + Fn(&'lua Lua, A) -> Result<R>;
@@ -292,11 +414,107 @@ pub trait UserDataMethods<'lua, T: UserData> {
     /// This is a version of [`add_meta_function`] that accepts a FnMut argument.
     ///
     /// [`add_meta_function`]: #method.add_meta_function
-    fn add_meta_function_mut<A, R, F>(&mut self, meta: MetaMethod, function: F)
+    fn add_meta_function_mut<S, A, R, F>(&mut self, meta: S, function: F)
     where
+        S: Into<MetaMethod>,
         A: FromLuaMulti<'lua>,
         R: ToLuaMulti<'lua>,
         F: 'static + MaybeSend + FnMut(&'lua Lua, A) -> Result<R>;
+
+    //
+    // Below are internal methods used in generated code
+    //
+
+    #[doc(hidden)]
+    fn add_callback(&mut self, _name: Vec<u8>, _callback: Callback<'lua, 'static>) {}
+
+    #[doc(hidden)]
+    #[cfg(feature = "async")]
+    fn add_async_callback(&mut self, _name: Vec<u8>, _callback: AsyncCallback<'lua, 'static>) {}
+
+    #[doc(hidden)]
+    fn add_meta_callback(&mut self, _meta: MetaMethod, _callback: Callback<'lua, 'static>) {}
+}
+
+/// Field registry for [`UserData`] implementors.
+///
+/// [`UserData`]: trait.UserData.html
+pub trait UserDataFields<'lua, T: UserData> {
+    /// Add a regular field getter as a method which accepts a `&T` as the parameter.
+    ///
+    /// Regular field getters are implemented by overriding the `__index` metamethod and returning the
+    /// accessed field. This allows them to be used with the expected `userdata.field` syntax.
+    ///
+    /// If `add_meta_method` is used to set the `__index` metamethod, the `__index` metamethod will
+    /// be used as a fall-back if no regular field or method are found.
+    fn add_field_method_get<S, R, M>(&mut self, name: &S, method: M)
+    where
+        S: AsRef<[u8]> + ?Sized,
+        R: ToLua<'lua>,
+        M: 'static + MaybeSend + Fn(&'lua Lua, &T) -> Result<R>;
+
+    /// Add a regular field setter as a method which accepts a `&mut T` as the first parameter.
+    ///
+    /// Regular field setters are implemented by overriding the `__newindex` metamethod and setting the
+    /// accessed field. This allows them to be used with the expected `userdata.field = value` syntax.
+    ///
+    /// If `add_meta_method` is used to set the `__newindex` metamethod, the `__newindex` metamethod will
+    /// be used as a fall-back if no regular field is found.
+    fn add_field_method_set<S, A, M>(&mut self, name: &S, method: M)
+    where
+        S: AsRef<[u8]> + ?Sized,
+        A: FromLua<'lua>,
+        M: 'static + MaybeSend + FnMut(&'lua Lua, &mut T, A) -> Result<()>;
+
+    /// Add a regular field getter as a function which accepts a generic [`AnyUserData`] of type `T`
+    /// argument.
+    ///
+    /// Prefer to use [`add_field_method_get`] as it is easier to use.
+    ///
+    /// [`AnyUserData`]: struct.AnyUserData.html
+    /// [`add_field_method_get`]: #method.add_field_method_get
+    fn add_field_function_get<S, R, F>(&mut self, name: &S, function: F)
+    where
+        S: AsRef<[u8]> + ?Sized,
+        R: ToLua<'lua>,
+        F: 'static + MaybeSend + Fn(&'lua Lua, AnyUserData<'lua>) -> Result<R>;
+
+    /// Add a regular field setter as a function which accepts a generic [`AnyUserData`] of type `T`
+    /// first argument.
+    ///
+    /// Prefer to use [`add_field_method_set`] as it is easier to use.
+    ///
+    /// [`AnyUserData`]: struct.AnyUserData.html
+    /// [`add_field_method_set`]: #method.add_field_method_set
+    fn add_field_function_set<S, A, F>(&mut self, name: &S, function: F)
+    where
+        S: AsRef<[u8]> + ?Sized,
+        A: FromLua<'lua>,
+        F: 'static + MaybeSend + FnMut(&'lua Lua, AnyUserData<'lua>, A) -> Result<()>;
+
+    /// Add a metamethod value computed from `f`.
+    ///
+    /// This will initialize the metamethod value from `f` on `UserData` creation.
+    ///
+    /// # Note
+    ///
+    /// `mlua` will trigger an error on an attempt to define a protected metamethod,
+    /// like `__gc` or `__metatable`.
+    fn add_meta_field_with<S, R, F>(&mut self, meta: S, f: F)
+    where
+        S: Into<MetaMethod>,
+        F: 'static + MaybeSend + Fn(&'lua Lua) -> Result<R>,
+        R: ToLua<'lua>;
+
+    //
+    // Below are internal methods used in generated code
+    //
+
+    #[doc(hidden)]
+    fn add_field_getter(&mut self, _name: Vec<u8>, _callback: Callback<'lua, 'static>) {}
+
+    #[doc(hidden)]
+    fn add_field_setter(&mut self, _name: Vec<u8>, _callback: Callback<'lua, 'static>) {}
 }
 
 /// Trait for custom userdata types.
@@ -322,21 +540,21 @@ pub trait UserDataMethods<'lua, T: UserData> {
 /// # }
 /// ```
 ///
-/// Custom methods and operators can be provided by implementing `add_methods` (refer to
-/// [`UserDataMethods`] for more information):
+/// Custom fields, methods and operators can be provided by implementing `add_fields` or `add_methods`
+/// (refer to [`UserDataFields`] and [`UserDataMethods`] for more information):
 ///
 /// ```
-/// # use mlua::{Lua, MetaMethod, Result, UserData, UserDataMethods};
+/// # use mlua::{Lua, MetaMethod, Result, UserData, UserDataFields, UserDataMethods};
 /// # fn main() -> Result<()> {
 /// # let lua = Lua::new();
 /// struct MyUserData(i32);
 ///
 /// impl UserData for MyUserData {
-///     fn add_methods<'lua, M: UserDataMethods<'lua, Self>>(methods: &mut M) {
-///         methods.add_method("get", |_, this, _: ()| {
-///             Ok(this.0)
-///         });
+///     fn add_fields<'lua, F: UserDataFields<'lua, Self>>(fields: &mut F) {
+///         fields.add_field_method_get("val", |_, this| Ok(this.0));
+///     }
 ///
+///     fn add_methods<'lua, M: UserDataMethods<'lua, Self>>(methods: &mut M) {
 ///         methods.add_method_mut("add", |_, this, value: i32| {
 ///             this.0 += value;
 ///             Ok(())
@@ -351,9 +569,9 @@ pub trait UserDataMethods<'lua, T: UserData> {
 /// lua.globals().set("myobject", MyUserData(123))?;
 ///
 /// lua.load(r#"
-///     assert(myobject:get() == 123)
+///     assert(myobject.val == 123)
 ///     myobject:add(7)
-///     assert(myobject:get() == 130)
+///     assert(myobject.val == 130)
 ///     assert(myobject + 10 == 140)
 /// "#).exec()?;
 /// # Ok(())
@@ -362,18 +580,158 @@ pub trait UserDataMethods<'lua, T: UserData> {
 ///
 /// [`ToLua`]: trait.ToLua.html
 /// [`FromLua`]: trait.FromLua.html
+/// [`UserDataFields`]: trait.UserDataFields.html
 /// [`UserDataMethods`]: trait.UserDataMethods.html
 pub trait UserData: Sized {
+    /// Adds custom fields specific to this userdata.
+    fn add_fields<'lua, F: UserDataFields<'lua, Self>>(_fields: &mut F) {}
+
     /// Adds custom methods and operators specific to this userdata.
     fn add_methods<'lua, M: UserDataMethods<'lua, Self>>(_methods: &mut M) {}
 }
 
+// Wraps UserData in a way to always implement `serde::Serialize` trait.
+pub(crate) struct UserDataCell<T>(RefCell<UserDataWrapped<T>>);
+
+impl<T> UserDataCell<T> {
+    pub(crate) fn new(data: T) -> Self {
+        UserDataCell(RefCell::new(UserDataWrapped::new(data)))
+    }
+
+    #[cfg(feature = "serialize")]
+    pub(crate) fn new_ser(data: T) -> Self
+    where
+        T: 'static + Serialize,
+    {
+        UserDataCell(RefCell::new(UserDataWrapped::new_ser(data)))
+    }
+
+    // Immutably borrows the wrapped value.
+    fn try_borrow(&self) -> Result<UserDataRef<T>> {
+        self.0
+            .try_borrow()
+            .map(|r| UserDataRef(UserDataRefInner::Ref(r)))
+            .map_err(|_| Error::UserDataBorrowError)
+    }
+
+    // Mutably borrows the wrapped value.
+    fn try_borrow_mut(&self) -> Result<UserDataRefMut<T>> {
+        self.0
+            .try_borrow_mut()
+            .map(|r| UserDataRefMut(UserDataRefMutInner::Ref(r)))
+            .map_err(|_| Error::UserDataBorrowMutError)
+    }
+}
+
+#[cfg(feature = "serialize")]
+impl Serialize for UserDataCell<()> {
+    fn serialize<S>(&self, serializer: S) -> StdResult<S::Ok, S::Error>
+    where
+        S: Serializer,
+    {
+        let ser = self
+            .0
+            .try_borrow()
+            .map_err(|_| ser::Error::custom(Error::UserDataBorrowError))?
+            .ser;
+        unsafe { (&*ser).serialize(serializer) }
+    }
+}
+
+/// A wrapper type for an immutably borrowed value from an `AnyUserData`.
+pub struct UserDataRef<'a, T>(UserDataRefInner<'a, T>);
+
+enum UserDataRefInner<'a, T> {
+    Ref(Ref<'a, UserDataWrapped<T>>),
+}
+
+/// A wrapper type for a mutably borrowed value from an `AnyUserData`.
+pub struct UserDataRefMut<'a, T>(UserDataRefMutInner<'a, T>);
+
+enum UserDataRefMutInner<'a, T> {
+    Ref(RefMut<'a, UserDataWrapped<T>>),
+}
+
+impl<T> Deref for UserDataRef<'_, T> {
+    type Target = T;
+
+    fn deref(&self) -> &Self::Target {
+        match &self.0 {
+            UserDataRefInner::Ref(x) => &*x,
+        }
+    }
+}
+
+impl<T> Deref for UserDataRefMut<'_, T> {
+    type Target = T;
+
+    fn deref(&self) -> &Self::Target {
+        match &self.0 {
+            UserDataRefMutInner::Ref(x) => &*x,
+        }
+    }
+}
+
+impl<T> DerefMut for UserDataRefMut<'_, T> {
+    fn deref_mut(&mut self) -> &mut Self::Target {
+        match &mut self.0 {
+            UserDataRefMutInner::Ref(x) => &mut *x,
+        }
+    }
+}
+
+impl<T: fmt::Debug> fmt::Debug for UserDataRef<'_, T> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        fmt::Debug::fmt(&*self as &T, f)
+    }
+}
+
+impl<T: fmt::Debug> fmt::Debug for UserDataRefMut<'_, T> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        fmt::Debug::fmt(&*self as &T, f)
+    }
+}
+
+impl<T: fmt::Display> fmt::Display for UserDataRef<'_, T> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        fmt::Display::fmt(&*self as &T, f)
+    }
+}
+
+impl<T: fmt::Display> fmt::Display for UserDataRefMut<'_, T> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        fmt::Display::fmt(&*self as &T, f)
+    }
+}
+
 pub(crate) struct UserDataWrapped<T> {
     pub(crate) data: *mut T,
     #[cfg(feature = "serialize")]
     ser: *mut dyn erased_serde::Serialize,
 }
 
+impl<T> UserDataWrapped<T> {
+    fn new(data: T) -> Self {
+        UserDataWrapped {
+            data: Box::into_raw(Box::new(data)),
+            #[cfg(feature = "serialize")]
+            ser: Box::into_raw(Box::new(UserDataSerializeError)),
+        }
+    }
+
+    #[cfg(feature = "serialize")]
+    fn new_ser(data: T) -> Self
+    where
+        T: 'static + Serialize,
+    {
+        let data_raw = Box::into_raw(Box::new(data));
+        UserDataWrapped {
+            data: data_raw,
+            ser: data_raw,
+        }
+    }
+}
+
 impl<T> Drop for UserDataWrapped<T> {
     fn drop(&mut self) {
         unsafe {
@@ -386,42 +744,22 @@ impl<T> Drop for UserDataWrapped<T> {
     }
 }
 
-impl<T> UserDataWrapped<T> {
-    pub(crate) fn new(data: T) -> Self {
-        UserDataWrapped {
-            data: Box::into_raw(Box::new(data)),
-            #[cfg(feature = "serialize")]
-            ser: Box::into_raw(Box::new(UserDataSerializeError)),
-        }
-    }
+impl<T> Deref for UserDataWrapped<T> {
+    type Target = T;
 
-    #[cfg(feature = "serialize")]
-    pub(crate) fn new_ser(data: T) -> Self
-    where
-        T: 'static + Serialize,
-    {
-        let data_raw = Box::into_raw(Box::new(data));
-        UserDataWrapped {
-            data: data_raw,
-            ser: data_raw,
-        }
-    }
-}
-
-impl<T> AsRef<T> for UserDataWrapped<T> {
-    fn as_ref(&self) -> &T {
+    fn deref(&self) -> &Self::Target {
         unsafe { &*self.data }
     }
 }
 
-impl<T> AsMut<T> for UserDataWrapped<T> {
-    fn as_mut(&mut self) -> &mut T {
+impl<T> DerefMut for UserDataWrapped<T> {
+    fn deref_mut(&mut self) -> &mut Self::Target {
         unsafe { &mut *self.data }
     }
 }
 
 #[cfg(feature = "serialize")]
-pub(crate) struct UserDataSerializeError;
+struct UserDataSerializeError;
 
 #[cfg(feature = "serialize")]
 impl Serialize for UserDataSerializeError {
@@ -468,26 +806,18 @@ impl<'lua> AnyUserData<'lua> {
     ///
     /// Returns a `UserDataBorrowError` if the userdata is already mutably borrowed. Returns a
     /// `UserDataTypeMismatch` if the userdata is not of type `T`.
-    pub fn borrow<T: 'static + UserData>(&self) -> Result<Ref<T>> {
-        self.inspect(|cell| {
-            let cell_ref = cell.try_borrow().map_err(|_| Error::UserDataBorrowError)?;
-            Ok(Ref::map(cell_ref, |x| unsafe { &*x.data }))
-        })
+    pub fn borrow<T: 'static + UserData>(&self) -> Result<UserDataRef<T>> {
+        self.inspect(|cell| cell.try_borrow())
     }
 
     /// Borrow this userdata mutably if it is of type `T`.
     ///
     /// # Errors
     ///
-    /// Returns a `UserDataBorrowMutError` if the userdata is already borrowed. Returns a
-    /// `UserDataTypeMismatch` if the userdata is not of type `T`.
-    pub fn borrow_mut<T: 'static + UserData>(&self) -> Result<RefMut<T>> {
-        self.inspect(|cell| {
-            let cell_ref = cell
-                .try_borrow_mut()
-                .map_err(|_| Error::UserDataBorrowMutError)?;
-            Ok(RefMut::map(cell_ref, |x| unsafe { &mut *x.data }))
-        })
+    /// Returns a `UserDataBorrowMutError` if the userdata cannot be mutably borrowed.
+    /// Returns a `UserDataTypeMismatch` if the userdata is not of type `T`.
+    pub fn borrow_mut<T: 'static + UserData>(&self) -> Result<UserDataRefMut<T>> {
+        self.inspect(|cell| cell.try_borrow_mut())
     }
 
     /// Sets an associated value to this `AnyUserData`.
@@ -500,19 +830,21 @@ impl<'lua> AnyUserData<'lua> {
         let lua = self.0.lua;
         #[cfg(any(feature = "lua52", feature = "lua51", feature = "luajit"))]
         let v = {
-            // Lua 5.2/5.1 allows to store only a table. Then we will wrap the value.
-            let t = lua.create_table()?;
+            // Lua <= 5.2 allows to store only a table. Then we will wrap the value.
+            let t = lua.create_table_with_capacity(1, 0)?;
             t.raw_set(1, v)?;
-            crate::Value::Table(t)
+            Value::Table(t)
         };
         #[cfg(any(feature = "lua54", feature = "lua53"))]
         let v = v.to_lua(lua)?;
         unsafe {
             let _sg = StackGuard::new(lua.state);
-            assert_stack(lua.state, 2);
-            lua.push_ref(&self.0);
+            check_stack(lua.state, 3)?;
+
+            lua.push_userdata_ref(&self.0, false)?;
             lua.push_value(v)?;
             ffi::lua_setuservalue(lua.state, -2);
+
             Ok(())
         }
     }
@@ -526,57 +858,54 @@ impl<'lua> AnyUserData<'lua> {
         let lua = self.0.lua;
         let res = unsafe {
             let _sg = StackGuard::new(lua.state);
-            assert_stack(lua.state, 3);
-            lua.push_ref(&self.0);
+            check_stack(lua.state, 3)?;
+
+            lua.push_userdata_ref(&self.0, false)?;
             ffi::lua_getuservalue(lua.state, -1);
             lua.pop_value()
         };
         #[cfg(any(feature = "lua52", feature = "lua51", feature = "luajit"))]
-        return crate::Table::from_lua(res, lua)?.get(1);
+        return match <Option<Table>>::from_lua(res, lua)? {
+            Some(t) => t.get(1),
+            None => V::from_lua(Value::Nil, lua),
+        };
         #[cfg(any(feature = "lua54", feature = "lua53"))]
         V::from_lua(res, lua)
     }
 
-    /// Checks for a metamethod in this `AnyUserData`
-    pub fn has_metamethod(&self, method: MetaMethod) -> Result<bool> {
-        match self.get_metatable() {
-            Ok(mt) => {
-                let name = self.0.lua.create_string(method.name())?;
-                if let Value::Nil = mt.raw_get(name)? {
-                    Ok(false)
-                } else {
-                    Ok(true)
-                }
-            }
-            Err(Error::UserDataTypeMismatch) => Ok(false),
-            Err(e) => Err(e),
-        }
+    /// Returns a metatable of this `UserData`.
+    ///
+    /// Returned [`UserDataMetatable`] object wraps the original metatable and
+    /// provides safe access to its methods.
+    ///
+    /// For `T: UserData + 'static` returned metatable is shared among all instances of type `T`.
+    ///
+    /// [`UserDataMetatable`]: struct.UserDataMetatable.html
+    pub fn get_metatable(&self) -> Result<UserDataMetatable<'lua>> {
+        self.get_raw_metatable().map(UserDataMetatable)
     }
 
-    fn get_metatable(&self) -> Result<Table<'lua>> {
+    fn get_raw_metatable(&self) -> Result<Table<'lua>> {
         unsafe {
             let lua = self.0.lua;
             let _sg = StackGuard::new(lua.state);
-            assert_stack(lua.state, 3);
-
-            lua.push_ref(&self.0);
-
-            if ffi::lua_getmetatable(lua.state, -1) == 0 {
-                return Err(Error::UserDataTypeMismatch);
-            }
+            check_stack(lua.state, 3)?;
 
+            lua.push_userdata_ref(&self.0, false)?;
+            ffi::lua_getmetatable(lua.state, -1); // Checked that non-empty on the previous call
             Ok(Table(lua.pop_ref()))
         }
     }
 
     pub(crate) fn equals<T: AsRef<Self>>(&self, other: T) -> Result<bool> {
         let other = other.as_ref();
+        // Uses lua_rawequal() under the hood
         if self == other {
             return Ok(true);
         }
 
-        let mt = self.get_metatable()?;
-        if mt != other.get_metatable()? {
+        let mt = self.get_raw_metatable()?;
+        if mt != other.get_raw_metatable()? {
             return Ok(false);
         }
 
@@ -589,39 +918,52 @@ impl<'lua> AnyUserData<'lua> {
         Ok(false)
     }
 
+    pub(crate) fn type_id(&self) -> Result<TypeId> {
+        let lua = self.0.lua;
+        unsafe {
+            let _sg = StackGuard::new(lua.state);
+            check_stack(lua.state, 5)?;
+
+            // Push userdata with metatable
+            lua.push_userdata_ref(&self.0, true)?;
+
+            // Get the special `__mlua_type_id`
+            push_string(lua.state, "__mlua_type_id")?;
+            if ffi::lua_rawget(lua.state, -2) != ffi::LUA_TUSERDATA {
+                return Err(Error::UserDataTypeMismatch);
+            }
+
+            Ok(*(ffi::lua_touserdata(lua.state, -1) as *const TypeId))
+        }
+    }
+
     fn inspect<'a, T, R, F>(&'a self, func: F) -> Result<R>
     where
         T: 'static + UserData,
         F: FnOnce(&'a UserDataCell<T>) -> Result<R>,
     {
+        let lua = self.0.lua;
         unsafe {
-            let lua = self.0.lua;
             let _sg = StackGuard::new(lua.state);
-            assert_stack(lua.state, 3);
+            check_stack(lua.state, 3)?;
 
             lua.push_ref(&self.0);
-
             if ffi::lua_getmetatable(lua.state, -1) == 0 {
-                Err(Error::UserDataTypeMismatch)
-            } else {
-                ffi::lua_rawgeti(
-                    lua.state,
-                    ffi::LUA_REGISTRYINDEX,
-                    lua.userdata_metatable::<T>()? as ffi::lua_Integer,
-                );
+                return Err(Error::UserDataTypeMismatch);
+            }
+            lua.push_userdata_metatable::<T>()?;
 
-                if ffi::lua_rawequal(lua.state, -1, -2) == 0 {
-                    // Maybe UserData destructed?
-                    ffi::lua_pop(lua.state, 1);
-                    get_destructed_userdata_metatable(lua.state);
-                    if ffi::lua_rawequal(lua.state, -1, -2) == 1 {
-                        Err(Error::UserDataDestructed)
-                    } else {
-                        Err(Error::UserDataTypeMismatch)
-                    }
+            if ffi::lua_rawequal(lua.state, -1, -2) == 0 {
+                // Maybe UserData destructed?
+                ffi::lua_pop(lua.state, 1);
+                get_destructed_userdata_metatable(lua.state);
+                if ffi::lua_rawequal(lua.state, -1, -2) == 1 {
+                    Err(Error::UserDataDestructed)
                 } else {
-                    func(&*get_userdata::<UserDataCell<T>>(lua.state, -3))
+                    Err(Error::UserDataTypeMismatch)
                 }
+            } else {
+                func(&*get_userdata::<UserDataCell<T>>(lua.state, -3))
             }
         }
     }
@@ -640,23 +982,95 @@ impl<'lua> AsRef<AnyUserData<'lua>> for AnyUserData<'lua> {
     }
 }
 
+/// Handle to a `UserData` metatable.
+#[derive(Clone, Debug)]
+pub struct UserDataMetatable<'lua>(pub(crate) Table<'lua>);
+
+impl<'lua> UserDataMetatable<'lua> {
+    /// Gets the value associated to `key` from the metatable.
+    ///
+    /// If no value is associated to `key`, returns the `Nil` value.
+    /// Access to restricted metamethods such as `__gc` or `__metatable` will cause an error.
+    pub fn get<K: Into<MetaMethod>, V: FromLua<'lua>>(&self, key: K) -> Result<V> {
+        self.0.raw_get(key.into().validate()?.name())
+    }
+
+    /// Sets a key-value pair in the metatable.
+    ///
+    /// If the value is `Nil`, this will effectively remove the `key`.
+    /// Access to restricted metamethods such as `__gc` or `__metatable` will cause an error.
+    /// Setting `__index` or `__newindex` metamethods is also restricted because their values are cached
+    /// for `mlua` internal usage.
+    pub fn set<K: Into<MetaMethod>, V: ToLua<'lua>>(&self, key: K, value: V) -> Result<()> {
+        let key = key.into().validate()?;
+        // `__index` and `__newindex` cannot be changed in runtime, because values are cached
+        if key == MetaMethod::Index || key == MetaMethod::NewIndex {
+            return Err(Error::MetaMethodRestricted(key.to_string()));
+        }
+        self.0.raw_set(key.name(), value)
+    }
+
+    /// Checks whether the metatable contains a non-nil value for `key`.
+    pub fn contains<K: Into<MetaMethod>>(&self, key: K) -> Result<bool> {
+        self.0.contains_key(key.into().validate()?.name())
+    }
+
+    /// Consumes this metatable and returns an iterator over the pairs of the metatable.
+    ///
+    /// The pairs are wrapped in a [`Result`], since they are lazily converted to `V` type.
+    ///
+    /// [`Result`]: type.Result.html
+    pub fn pairs<V: FromLua<'lua>>(self) -> UserDataMetatablePairs<'lua, V> {
+        UserDataMetatablePairs(self.0.pairs())
+    }
+}
+
+/// An iterator over the pairs of a [`UserData`] metatable.
+///
+/// It skips restricted metamethods, such as `__gc` or `__metatable`.
+///
+/// This struct is created by the [`UserDataMetatable::pairs`] method.
+///
+/// [`UserData`]: trait.UserData.html
+/// [`UserDataMetatable::pairs`]: struct.UserDataMetatable.html#method.pairs
+pub struct UserDataMetatablePairs<'lua, V>(TablePairs<'lua, StdString, V>);
+
+impl<'lua, V> Iterator for UserDataMetatablePairs<'lua, V>
+where
+    V: FromLua<'lua>,
+{
+    type Item = Result<(MetaMethod, V)>;
+
+    fn next(&mut self) -> Option<Self::Item> {
+        loop {
+            match self.0.next()? {
+                Ok((key, value)) => {
+                    // Skip restricted metamethods
+                    if let Ok(metamethod) = MetaMethod::from(key).validate() {
+                        break Some(Ok((metamethod, value)));
+                    }
+                }
+                Err(e) => break Some(Err(e)),
+            }
+        }
+    }
+}
+
 #[cfg(feature = "serialize")]
 impl<'lua> Serialize for AnyUserData<'lua> {
     fn serialize<S>(&self, serializer: S) -> StdResult<S::Ok, S::Error>
     where
         S: Serializer,
     {
-        let f = || unsafe {
+        unsafe {
             let lua = self.0.lua;
             let _sg = StackGuard::new(lua.state);
-            assert_stack(lua.state, 2);
+            check_stack(lua.state, 3).map_err(ser::Error::custom)?;
 
-            lua.push_userdata_ref(&self.0)?;
+            lua.push_userdata_ref(&self.0, false)
+                .map_err(ser::Error::custom)?;
             let ud = &*get_userdata::<UserDataCell<()>>(lua.state, -1);
-            (*ud.try_borrow().map_err(|_| Error::UserDataBorrowError)?.ser)
-                .serialize(serializer)
-                .map_err(|err| Error::SerializeError(err.to_string()))
-        };
-        f().map_err(ser::Error::custom)
+            ud.serialize(serializer)
+        }
     }
 }
