@@ -1,11 +1,14 @@
-use std::sync::Arc;
+use std::sync::{Arc, Mutex, RwLock};
+
+#[cfg(not(feature = "send"))]
+use std::{cell::RefCell, rc::Rc};
 
 #[cfg(feature = "lua54")]
 use std::sync::atomic::{AtomicI64, Ordering};
 
 use mlua::{
-    AnyUserData, ExternalError, Function, Lua, MetaMethod, Result, String, UserData,
-    UserDataMethods, Value,
+    AnyUserData, Error, ExternalError, Function, Lua, MetaMethod, Nil, Result, String, UserData,
+    UserDataFields, UserDataMethods, Value,
 };
 
 #[test]
@@ -155,10 +158,10 @@ fn test_metamethods() -> Result<()> {
     assert!(userdata2.equals(userdata3)?);
 
     let userdata1: AnyUserData = globals.get("userdata1")?;
-    assert!(userdata1.has_metamethod(MetaMethod::Add)?);
-    assert!(userdata1.has_metamethod(MetaMethod::Sub)?);
-    assert!(userdata1.has_metamethod(MetaMethod::Index)?);
-    assert!(!userdata1.has_metamethod(MetaMethod::Pow)?);
+    assert!(userdata1.get_metatable()?.contains(MetaMethod::Add)?);
+    assert!(userdata1.get_metatable()?.contains(MetaMethod::Sub)?);
+    assert!(userdata1.get_metatable()?.contains(MetaMethod::Index)?);
+    assert!(!userdata1.get_metatable()?.contains(MetaMethod::Pow)?);
 
     Ok(())
 }
@@ -250,7 +253,7 @@ fn test_gc_userdata() -> Result<()> {
 }
 
 #[test]
-fn detroys_userdata() -> Result<()> {
+fn test_destroy_userdata() -> Result<()> {
     struct MyUserdata(Arc<()>);
 
     impl UserData for MyUserdata {}
@@ -272,7 +275,7 @@ fn detroys_userdata() -> Result<()> {
 }
 
 #[test]
-fn user_value() -> Result<()> {
+fn test_user_value() -> Result<()> {
     struct MyUserData;
 
     impl UserData for MyUserData {}
@@ -295,7 +298,7 @@ fn test_functions() -> Result<()> {
             methods.add_function("get_value", |_, ud: AnyUserData| {
                 Ok(ud.borrow::<MyUserData>()?.0)
             });
-            methods.add_function("set_value", |_, (ud, value): (AnyUserData, i64)| {
+            methods.add_function_mut("set_value", |_, (ud, value): (AnyUserData, i64)| {
                 ud.borrow_mut::<MyUserData>()?.0 = value;
                 Ok(())
             });
@@ -335,3 +338,194 @@ fn test_functions() -> Result<()> {
 
     Ok(())
 }
+
+#[test]
+fn test_fields() -> Result<()> {
+    #[derive(Copy, Clone)]
+    struct MyUserData(i64);
+
+    impl UserData for MyUserData {
+        fn add_fields<'lua, F: UserDataFields<'lua, Self>>(fields: &mut F) {
+            fields.add_field_method_get("val", |_, data| Ok(data.0));
+            fields.add_field_method_set("val", |_, data, val| {
+                data.0 = val;
+                Ok(())
+            });
+
+            // Use userdata "uservalue" storage
+            fields.add_field_function_get("uval", |_, ud| ud.get_user_value::<Option<String>>());
+            fields
+                .add_field_function_set("uval", |_, ud, s| ud.set_user_value::<Option<String>>(s));
+
+            fields.add_meta_field_with(MetaMethod::Index, |lua| {
+                let index = lua.create_table()?;
+                index.set("f", 321)?;
+                Ok(index)
+            });
+            fields.add_meta_field_with(MetaMethod::NewIndex, |lua| {
+                lua.create_function(|lua, (_, field, val): (AnyUserData, String, Value)| {
+                    lua.globals().set(field, val)?;
+                    Ok(())
+                })
+            })
+        }
+    }
+
+    let lua = Lua::new();
+    let globals = lua.globals();
+    globals.set("ud", MyUserData(7))?;
+    lua.load(
+        r#"
+        assert(ud.val == 7)
+        ud.val = 10
+        assert(ud.val == 10)
+
+        assert(ud.uval == nil)
+        ud.uval = "hello"
+        assert(ud.uval == "hello")
+
+        assert(ud.f == 321)
+
+        ud.unknown = 789
+        assert(unknown == 789)
+    "#,
+    )
+    .exec()?;
+
+    Ok(())
+}
+
+#[test]
+fn test_metatable() -> Result<()> {
+    #[derive(Copy, Clone)]
+    struct MyUserData(i64);
+
+    impl UserData for MyUserData {
+        fn add_fields<'lua, F: UserDataFields<'lua, Self>>(fields: &mut F) {
+            fields.add_meta_field_with("__type_name", |_| Ok("MyUserData"));
+        }
+
+        fn add_methods<'lua, M: UserDataMethods<'lua, Self>>(methods: &mut M) {
+            methods.add_function("my_type_name", |_, data: AnyUserData| {
+                let metatable = data.get_metatable()?;
+                metatable.get::<_, String>("__type_name")
+            });
+        }
+    }
+
+    let lua = Lua::new();
+    let globals = lua.globals();
+    globals.set("ud", MyUserData(7))?;
+    lua.load(
+        r#"
+        assert(ud:my_type_name() == "MyUserData")
+    "#,
+    )
+    .exec()?;
+
+    let ud: AnyUserData = globals.get("ud")?;
+    let metatable = ud.get_metatable()?;
+
+    match metatable.get::<_, Value>("__gc") {
+        Ok(_) => panic!("expected MetaMethodRestricted, got no error"),
+        Err(Error::MetaMethodRestricted(_)) => {}
+        Err(e) => panic!("expected MetaMethodRestricted, got {:?}", e),
+    }
+
+    match metatable.set(MetaMethod::Index, Nil) {
+        Ok(_) => panic!("expected MetaMethodRestricted, got no error"),
+        Err(Error::MetaMethodRestricted(_)) => {}
+        Err(e) => panic!("expected MetaMethodRestricted, got {:?}", e),
+    }
+
+    let mut methods = metatable
+        .pairs()
+        .into_iter()
+        .map(|kv: Result<(_, Value)>| Ok(kv?.0))
+        .collect::<Result<Vec<_>>>()?;
+    methods.sort_by_cached_key(|k| k.name().to_owned());
+    assert_eq!(methods, vec![MetaMethod::Index, "__type_name".into()]);
+
+    #[derive(Copy, Clone)]
+    struct MyUserData2(i64);
+
+    impl UserData for MyUserData2 {
+        fn add_fields<'lua, F: UserDataFields<'lua, Self>>(fields: &mut F) {
+            fields.add_meta_field_with("__index", |_| Ok(1));
+        }
+    }
+
+    match lua.create_userdata(MyUserData2(1)) {
+        Ok(_) => panic!("expected MetaMethodTypeError, got no error"),
+        Err(Error::MetaMethodTypeError { .. }) => {}
+        Err(e) => panic!("expected MetaMethodTypeError, got {:?}", e),
+    }
+
+    Ok(())
+}
+
+#[test]
+fn test_userdata_wrapped() -> Result<()> {
+    struct MyUserData(i64);
+
+    impl UserData for MyUserData {
+        fn add_fields<'lua, F: UserDataFields<'lua, Self>>(fields: &mut F) {
+            fields.add_field_method_get("data", |_, this| Ok(this.0));
+            fields.add_field_method_set("data", |_, this, val| {
+                this.0 = val;
+                Ok(())
+            })
+        }
+    }
+
+    let lua = Lua::new();
+    let globals = lua.globals();
+
+    #[cfg(not(feature = "send"))]
+    {
+        let ud1 = Rc::new(RefCell::new(MyUserData(1)));
+        globals.set("rc_refcell_ud", ud1.clone())?;
+        lua.load(
+            r#"
+            rc_refcell_ud.data = rc_refcell_ud.data + 1
+            assert(rc_refcell_ud.data == 2)
+        "#,
+        )
+        .exec()?;
+        assert_eq!(ud1.borrow().0, 2);
+        globals.set("rc_refcell_ud", Nil)?;
+        lua.gc_collect()?;
+        assert_eq!(Rc::strong_count(&ud1), 1);
+    }
+
+    let ud2 = Arc::new(Mutex::new(MyUserData(2)));
+    globals.set("arc_mutex_ud", ud2.clone())?;
+    lua.load(
+        r#"
+        arc_mutex_ud.data = arc_mutex_ud.data + 1
+        assert(arc_mutex_ud.data == 3)
+    "#,
+    )
+    .exec()?;
+    assert_eq!(ud2.lock().unwrap().0, 3);
+
+    let ud3 = Arc::new(RwLock::new(MyUserData(3)));
+    globals.set("arc_rwlock_ud", ud3.clone())?;
+    lua.load(
+        r#"
+        arc_rwlock_ud.data = arc_rwlock_ud.data + 1
+        assert(arc_rwlock_ud.data == 4)
+    "#,
+    )
+    .exec()?;
+    assert_eq!(ud3.read().unwrap().0, 4);
+
+    // Test drop
+    globals.set("arc_mutex_ud", Nil)?;
+    globals.set("arc_rwlock_ud", Nil)?;
+    lua.gc_collect()?;
+    assert_eq!(Arc::strong_count(&ud2), 1);
+    assert_eq!(Arc::strong_count(&ud3), 1);
+
+    Ok(())
+}
